#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <poll.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <memcache.h>
#include <adns.h>
#include <time.h>
#include <sys/time.h>
#include <syslog.h>

struct udp_message {
	struct udp_message	*next;
	long	msgid;
	int	sent;
	char	*host;
};

struct udp_message	*head, *tail;

static struct pollfd	fds [2];
static struct memcache	*mc;

static	int	udp_port = -1, ttl = -1, nodaemon;
static	char	*bind_addr, *srv_addr, *key, *mc_list;
static	adns_state	adns;

void	init_memcached_connection (char *mc_list);
void	init_udp_connection (char *bind_addr, int udp_port);
void	init_adns (char *srv_addr);
void	loop (void);
char	*process_dns_message (adns_answer *ans);

void	usage (void)
{
	printf ("Usage is: host2cat -f [-h] -m mc_list -u udp_port -s ip_address[,ip_address]... -t ttl [-a bind_address]\n");
}

int	main (int argc, char **argv)
{
	int	opt;

	while ((opt = getopt (argc, argv, "hfm:u:s:t:k:a:")) != -1) {
		switch (opt) {
			case 'h':
				usage ();
				return 0;
			case 'f':
				nodaemon++;
				break;
			case 'm':
				mc_list = optarg;
				break;
			case 'u':
				udp_port = atoi (optarg);
				break;
			case 's':
				srv_addr = optarg;
				break;
			case 't':
				ttl = atoi (optarg);
				break;
			case 'k':
				key = optarg;
				break;
			case 'a':
				bind_addr = optarg;
				break;
			case '?':
				usage ();
				return 1;
				break;
		}
	}

	if (!mc_list || udp_port == -1 || !srv_addr || ttl == -1) {
		usage ();
		return 1;
	}

	init_memcached_connection (mc_list);
	init_udp_connection (bind_addr, udp_port);
	init_adns (srv_addr);
	if (!nodaemon)	daemon (0, 0);
	loop ();
	adns_finish (adns);
	return 0;
}

void	loop (void)
{
	char	udpbuf [2048];
	long	msgid;

	srandom (time (NULL));
	msgid = random ();
	fds [0].events = POLLIN;
	struct timeval timenow;

	while (1) {
		struct udp_message	*msg;
		int	n = 1, bytes, to = -1;

//		for (msg = head; msg; msg = msg->next)
//			printf ("msgid: %ld, sent: %d, host: %s\n", msg->msgid, msg->sent, msg->host);

		while (1) {
			struct udp_message	*pmsg = NULL;
			adns_query	qu = NULL;
			adns_answer	*ans;
			int	rc;
			char	*cats;

			rc = adns_check (adns, &qu, &ans, (void **) &pmsg);
//			printf ("adns_check returns %d\n", rc);
			if (rc) {
				if (rc != EAGAIN && rc != ESRCH)	fprintf (stderr, "adns_check returns %d\n", rc);
				break;
			}
			cats = process_dns_message (ans);
			free (ans);
			if (nodaemon)	printf ("cats for host %s: %s\n", pmsg->host, cats);

			for (msg = head; msg && msg != pmsg; msg = msg->next)
					;
			if (!msg) {
				fprintf (stderr, "Can't find message\n");
				continue;
			}

			if (cats) {
				rc = mc_add (mc, msg->host, strlen (msg->host), cats, strlen (cats) + 1, ttl, 0);
				if (rc)	fprintf (stderr, "Can't store categories '%s' at memcahed servers: %d\n", cats, rc);
				free (cats);
			}

			if (head == pmsg) {
				head = pmsg->next;
				if (!head)	tail = head;
			} else {
				for (msg = head; msg->next && msg->next != pmsg; msg = msg->next)
					;
				msg->next = pmsg->next;
				if (!msg->next)	tail = msg;
			}
			free (pmsg->host);
			free (pmsg);
		}

		adns_beforepoll (adns, fds + 1, &n, &to, NULL);

//		printf ("events0: %d, events1: %d\n", fds [0].events, fds [1].events);
		if ((n = poll (fds, n + 1, to)) <= 0) {
                        if (n < 0 && errno != EINTR) {
				fprintf (stderr, "poll () returns error: %s\n", strerror (errno));
				sleep (1);
			}
                        continue;
                }
//		printf ("revents0: %x, revents1: %x\n", fds [0].revents, fds [1].revents);

		if (fds [0].revents & POLLIN) {
			if ((bytes = read (fds [0].fd, udpbuf, sizeof udpbuf)) <= 0)
				fprintf (stderr, "Bad UDP message read: %d\n", bytes);
			else if (!(msg = calloc (1, sizeof *msg)))
				fprintf (stderr, "Can't allocate memory for UDP message structure\n");
			else if (!(msg->host = malloc (bytes))) {
				fprintf (stderr, "Can't allocate memory for UDP message host\n");
				free (msg);
			} else {
				adns_query	qu;

//				printf ("Got request: '%s'\n", udpbuf);
				msg->msgid = msgid++;
				bcopy (udpbuf, msg->host, bytes);
				if (tail) {
					tail->next = msg;
					tail = msg;
				} else
					head = tail = msg;
				adns_submit (adns, msg->host, adns_r_txt, adns_qf_owner, msg, &qu);
				gettimeofday(&timenow, NULL);
				syslog (LOG_INFO, "adns_submit %u %u", timenow.tv_sec, timenow.tv_usec);
			}
			n--;
		}
		adns_afterpoll (adns, fds + 1, n, NULL);
	}
}

void	init_memcached_connection (char *mc_list)
{
	int	nservers, retval;
	char	*p, *p1, *ep;

	if (!(mc = mc_new ())) {
		fprintf (stderr, "No memory for mc structure\n");
		exit (1);
	}

	ep = mc_list + strlen(mc_list);
	for (nservers = 0, p = mc_list; p < ep; p = p1 + 1) {
		// p points to beginning of server, p1 points to end of server
		p1 = strchr (p, ',');
		if (!p1)	p1 = p + strlen (p);

		if (p1 == p)
			fprintf (stderr, "Empty server in '%s' - ignored\n", mc_list);
		else if ((retval = mc_server_add5 (mc, p, p1 - p)))
			fprintf (stderr, "server add for server %.*s returns %d\n", p1 - p, p, retval);
		else {
			nservers++;
			fprintf (stderr, "server %.*s successfully added\n", p1 - p, p);
		}
	}

	if (!nservers) {
		fprintf (stderr, "No servers added\n");
		mc_free (mc);
		mc = NULL;
		exit (1);
	}
}

void init_udp_connection (char *bind_addr, int udp_port)
{
	struct sockaddr_in	addr;

	bzero (&addr, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_port = htons (udp_port);
	if (bind_addr)
		inet_aton (bind_addr, &addr.sin_addr);
	else
		addr.sin_addr.s_addr = htonl (INADDR_ANY);

	if ((fds [0].fd = socket (AF_INET, SOCK_DGRAM, 0)) == -1) {
		fprintf (stderr, "Can't open UDP socket: %s\n", strerror (errno));
		exit (1);
	}

	if (fcntl (fds [0].fd,  F_SETFL, O_NONBLOCK) == -1) {
		fprintf (stderr, "Can't issue fcntl on UDP socket: %s\n", strerror (errno));
		exit (1);
	}

	if (bind (fds [0].fd, &addr, sizeof addr) == - 1) {
		fprintf (stderr, "Can't bind UDP socket: %s\n", strerror (errno));
		exit (1);
	}

}

void	init_adns (char *srv_addr)
{
	char	buf [1024], *p, *ns;

	for (*buf = 0, p = strtok (srv_addr, ","); p; p = strtok (NULL, ",")) {
		if (!strlen (p)) {
			fprintf (stderr, "Empty nameserver address\n");
			continue;
		}

		if (inet_addr(p) == INADDR_NONE) {
			struct hostent *he;

			if ((he = gethostbyname(p)) == NULL) {
				fprintf(stderr, "%s: %s\n", p, hstrerror(h_errno));
				continue;
			}
			ns = inet_ntoa(*(struct in_addr *)he->h_addr);
		} else {
			ns = p;
		}

		sprintf (buf + strlen (buf), "nameserver %s\n", ns);
	}

	errno = adns_init_strcfg (&adns, adns_if_debug, stderr, buf);
//	errno = adns_init_logfn (&adns, adns_if_debug, buf, NULL, stderr);
	if (errno) {
		fprintf (stderr, "adns_init return error: %s\n", strerror (errno));
		exit (1);
	}
}

char	*process_dns_message (adns_answer *ans)
{
	int	i, len = 0;
	char	*cats, *p;

//	printf ("status: %d\n", ans->status);
//	printf ("nrrs: %d\n", ans->nrrs);
//	printf ("owner: %s\n", ans->owner);
	if (ans->status == adns_s_nxdomain || ans->status == adns_s_nodata)
		len = 1;
	else if (ans->status == adns_s_ok) {
		for (i = 0; i < ans->nrrs; i++)
			len += strlen (ans->rrs.manyistr [i]->str) + 1;
		if (!len)	len = 1;
	}
	else
		return NULL;

	cats = malloc (len);
	if (!cats) {
		fprintf (stderr, "Can't allocate memory for categories list\n");
		return NULL;
	}
	cats [len - 1] = 0;

	if (len == 1)	return cats;

	for (p = cats, i = 0; i < ans->nrrs; i++) {
		bcopy (ans->rrs.manyistr [i]->str, p, strlen (ans->rrs.manyistr [i]->str));
		p += strlen (ans->rrs.manyistr [i]->str) + 1;
		p [-1] = ':';
	}
	p [-1] = 0;
	for (p = strchr (cats, ','); p; p = strchr (p + 1, ','))
		*p = ':';
	return cats;
}
